"""
A daemon that cleans up tasks from the service queues when a service is disabled/deleted.

When a service is turned off by the orchestrator or deleted by the user, the service task queue needs to be
emptied. The status of all the services will be periodically checked and any service that is found to be
disabled or deleted for which a service queue exists, the dispatcher will be informed that the task(s)
had an error.
"""
import threading
import warnings
from typing import Optional

from assemblyline.common.constants import service_queue_name
from assemblyline.common.forge import get_service_queue
from assemblyline.common.isotime import now_as_iso
from assemblyline.odm.models.error import Error
from assemblyline.odm.models.service import Service
from assemblyline.remote.datatypes import retry_call
from assemblyline.remote.datatypes.queues.named import NamedQueue
from assemblyline_core.dispatching.client import DispatchClient
from assemblyline_core.server_base import CoreBase, ServiceStage

DAY = 60 * 60 * 24
TASK_DELETE_CHUNK = 10000


class Plumber(CoreBase):
    def __init__(self, logger=None, shutdown_timeout: Optional[float] = None, config=None,
                 redis=None, redis_persist=None, datastore=None, delay=60):
        super().__init__('assemblyline.plumber', logger, shutdown_timeout, config=config, redis=redis,
                         redis_persist=redis_persist, datastore=datastore)
        self.delay = float(delay)
        self.datastore.ds.switch_user("plumber")
        self.dispatch_client = DispatchClient(datastore=self.datastore, redis=self.redis,
                                              redis_persist=self.redis_persist, logger=self.log)

        self.flush_threads: dict[str, threading.Thread] = {}
        self.stop_signals: dict[str, threading.Event] = {}
        self.service_limit: dict[str, int] = {}

    def stop(self):
        for sig in self.stop_signals.values():
            sig.set()
        super().stop()

    def try_run(self):
        # Start a task cleanup thread
        tc_thread = threading.Thread(target=self.cleanup_old_tasks, daemon=True, name="datastore_task_cleanup")
        tc_thread.start()

        # Start a notification queue cleanup thread
        nq_thread = threading.Thread(target=self.cleanup_notification_queues, daemon=True,
                                     name="redis_notification_queue_cleanup")
        nq_thread.start()

        self.service_queue_plumbing()

    def service_queue_plumbing(self):
        # Get an initial list of all the service queues
        service_queues: dict[str, Optional[Service]]
        service_queues = {queue.decode('utf-8').lstrip('service-queue-'): None
                          for queue in self.redis.keys(service_queue_name('*'))}

        while self.running:
            # Reset the status of the service queues
            service_queues = {service_name: None for service_name in service_queues}

            # Update the service queue status based on current list of services
            for service in self.datastore.list_all_services(full=True):
                service_queues[service.name] = service

            for service_name, service in service_queues.items():
                # For disabled or othewise unavailable services purge the queue
                current_stage = self.get_service_stage(service_name, ServiceStage.Running)
                if not service or not service.enabled or current_stage != ServiceStage.Running:
                    while True:
                        task = self.dispatch_client.request_work('plumber', service_name=service_name,
                                                                 service_version='0', blocking=False)
                        if task is None:
                            break

                        error = Error(dict(
                            archive_ts=None,
                            created='NOW',
                            expiry_ts=now_as_iso(task.ttl * DAY) if task.ttl else None,
                            response=dict(
                                message='The service was disabled while processing this task.',
                                service_name=task.service_name,
                                service_version='0',
                                status='FAIL_NONRECOVERABLE',
                            ),
                            sha256=task.fileinfo.sha256,
                            type="TASK PRE-EMPTED",
                        ))

                        error_key = error.build_key(task=task)
                        self.dispatch_client.service_failed(task.sid, error_key, error)
                        self.heartbeat()

                # For services that are enabled but limited
                if not service or not service.enabled or service.max_queue_length == 0:
                    if service_name in self.stop_signals:
                        self.stop_signals[service_name].set()
                        self.service_limit.pop(service_name)
                        self.flush_threads.pop(service_name)
                elif service and service.enabled and service.max_queue_length > 0:
                    self.service_limit[service_name] = service.max_queue_length
                    thread = self.flush_threads.get(service_name)
                    if not thread or not thread.is_alive():
                        self.stop_signals[service_name] = threading.Event()
                        thread = threading.Thread(
                            target=self.watch_service, args=[service_name],
                            daemon=True, name=service_name)
                        self.flush_threads[service_name] = thread
                        thread.start()

            # Wait a while before checking status of all services again
            self.sleep_with_heartbeat(self.delay)

    def cleanup_notification_queues(self):
        self.log.info("Cleaning up notification queues for old messages...")
        while self.running:
            # Finding all possible notification queues
            keys = [k.decode() for k in retry_call(self.redis_persist.keys, "nq-*")]
            if not keys:
                self.log.info('There are no queues right now in the system')
            for k in keys:
                self.log.info(f'Checking for old message in queue: {k}')
                # Peek the first message of the queue
                q = NamedQueue(k, self.redis_persist)
                msg = q.peek_next()
                if msg:
                    current_time = now_as_iso(-1 * self.config.core.plumber.notification_queue_max_age)
                    task_time = msg.get('notify_time', None) or msg.get('ingest_time', None)

                    # If the message too old, cleanup
                    if task_time is None or task_time <= current_time:
                        self.log.warning(
                            f"Messages on queue {k} by "
                            f"{msg.get('submission', {}).get('params', {}).get('submitter', 'unknown')}"
                            f" are older then the maximum queue age ({task_time}), removing queue")
                        q.delete()
                    else:
                        self.log.info('All messages are recent enough')
                else:
                    self.log.info('There are no messages in the queue')

            # wait for next run
            self.sleep(self.config.core.plumber.notification_queue_interval)
        self.log.info("Done cleaning up notification queues")

    def cleanup_old_tasks(self):
        self.log.info("Cleaning up task index for old completed tasks...")
        while self.running:
            deleted = self.datastore.task_cleanup(deleteable_task_age=DAY, max_tasks=TASK_DELETE_CHUNK)
            if not deleted:
                self.sleep(self.delay)
            else:
                self.log.info(f"Cleaned up {deleted} tasks that were already completed")
        self.log.info("Done cleaning up task index")

    def watch_service(self, service_name):
        self.log.info(f"Watching {service_name} service queue...")
        service_queue = get_service_queue(service_name, self.redis)
        while self.running and not self.stop_signals[service_name].is_set():
            while service_queue.length() > self.service_limit[service_name]:
                task = self.dispatch_client.request_work('plumber', service_name=service_name,
                                                         service_version='0', blocking=False, low_priority=True)
                if task is None:
                    break

                error = Error(dict(
                    archive_ts=None,
                    created='NOW',
                    expiry_ts=now_as_iso(task.ttl * 24 * 60 * 60) if task.ttl else None,
                    response=dict(
                        message="Task canceled due to execesive queuing.",
                        service_name=task.service_name,
                        service_version='0',
                        status='FAIL_NONRECOVERABLE',
                    ),
                    sha256=task.fileinfo.sha256,
                    type="TASK PRE-EMPTED",
                ))

                error_key = error.build_key(task=task)
                self.dispatch_client.service_failed(task.sid, error_key, error)
            self.sleep(2)
        self.log.info(f"Done watching {service_name} service queue")


if __name__ == '__main__':
    with Plumber() as server:
        server.serve_forever()



import collections
import json
import os
import random
import time

import pytest

from assemblyline.common import forge
from assemblyline.odm.random_data import create_alerts, wipe_alerts, wipe_submissions, create_submission, create_badlists, create_safelists, create_workflows, wipe_badlist, wipe_safelist, wipe_workflows, create_signatures, wipe_signatures
from assemblyline_core.replay.creator.run import ReplayCreator
from assemblyline_core.replay.creator.run_worker import ReplayCreatorWorker
from assemblyline_core.replay.loader.run import ReplayLoader
from assemblyline_core.replay.loader.run_worker import ReplayLoaderWorker

NUM_ALERTS = 1
NUM_BADLIST_ITEMS = 1
NUM_SAFELIST_ITEMS = 1
NUM_SIGNATURES = 1
NUM_SUBMISSIONS = 1
NUM_WORKFLOWS = 1

all_alerts = []
all_submissions = []

data_collections = collections.defaultdict(list)
DATA_COLLECTION_NAMES = ["badlist", "safelist", "workflow"]


@pytest.fixture(scope='module')
def fs(config):
    return forge.get_filestore(config)


def recursive_extend(d, u):
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = recursive_extend(d.get(k, {}), v)
        else:
            if k not in d:
                d[k] = []
            d[k].extend(v)

    return d


@pytest.fixture(scope="module")
def datastore(request, datastore_connection, fs):
    wipe_alerts(datastore_connection)
    wipe_badlist(datastore_connection)
    wipe_safelist(datastore_connection)
    wipe_submissions(datastore_connection, fs)
    wipe_signatures(datastore_connection)
    wipe_workflows(datastore_connection)

    for _ in range(NUM_SUBMISSIONS):
        all_submissions.append(create_submission(datastore_connection, fs))
    create_alerts(datastore_connection, alert_count=NUM_ALERTS,
                  submission_list=all_submissions)
    create_badlists(datastore_connection, count=NUM_BADLIST_ITEMS)

    # Generate all signatures from testing set, but only keep what's being asked to limit to
    create_signatures(datastore_connection)
    data_collections["signature"] = \
        datastore_connection.signature.search("*", rows=NUM_SIGNATURES, fl="id,*")['items']
    wipe_signatures(datastore_connection)
    for sig in data_collections["signature"]:
        datastore_connection.signature.save(sig.id, sig)

    create_safelists(datastore_connection, count=NUM_SAFELIST_ITEMS)
    create_workflows(datastore_connection, count=NUM_WORKFLOWS)
    for alert in datastore_connection.alert.stream_search("id:*", fl="*"):
        all_alerts.append(alert)

    for collection in DATA_COLLECTION_NAMES:
        for i in getattr(datastore_connection, collection).stream_search("id:*", fl="id,*"):
            data_collections[collection].append(i)

    try:
        yield datastore_connection
    finally:
        wipe_alerts(datastore_connection)
        wipe_badlist(datastore_connection)
        wipe_safelist(datastore_connection)
        wipe_signatures(datastore_connection)
        wipe_submissions(datastore_connection, fs)
        wipe_workflows(datastore_connection)


@pytest.fixture(scope="module")
def creator():
    # Initialize Replay Creator
    replay_creator = ReplayCreator()
    replay_creator.running = True
    # Clear all queues
    for q in replay_creator.client.queues.values():
        q.delete()
    output_dir = replay_creator.replay_config.creator.output_filestore.replace('file://', '')
    if os.path.exists(output_dir):
        for f in os.listdir(output_dir):
            os.unlink(os.path.join(output_dir, f))

    return replay_creator


@pytest.fixture(scope="module")
def creator_worker():
    # Initialize Replay Creator worker
    replay_creator_worker = ReplayCreatorWorker()
    replay_creator_worker.running = True

    return replay_creator_worker


@pytest.fixture(scope="module")
def loader():
    # Initialize Replay Loader
    replay_loader = ReplayLoader()
    replay_loader.running = True
    replay_loader.client.queues['file'].delete()
    input_dir = replay_loader.replay_config.loader.input_directory
    if os.path.exists(input_dir):
        for f in os.listdir(input_dir):
            os.unlink(os.path.join(input_dir, f))

    return replay_loader


@pytest.fixture(scope="module")
def loader_worker():
    # Initialize Replay Loader worker
    replay_loader_worker = ReplayLoaderWorker()
    replay_loader_worker.running = True

    return replay_loader_worker


def test_replay_single_alert(config, datastore, creator, creator_worker, loader, loader_worker):
    output_dir = creator.replay_config.creator.output_filestore.replace('file://', '')
    input_dir = loader.replay_config.loader.input_directory

    # Make sure the alert get picked up by the creator
    alert = random.choice(all_alerts)
    alert.extended_scan = 'completed'
    alert.workflows_completed = True
    datastore.alert.save(alert.alert_id, alert)
    datastore.alert.commit()
    filename = os.path.join(output_dir, f'alert_{alert.alert_id}.al_bundle')

    # Test replay creator
    creator.client.setup_alert_input_queue(once=True)
    assert creator.client.queues['alert'].length() == 1
    assert creator.client.queues['alert'].peek_next()[
        'alert_id'] == alert.alert_id

    # Test replay creator worker
    creator_worker.process_alert(once=True)
    datastore.alert.commit()
    assert creator_worker.client.queues['alert'].length() == 0
    assert datastore.alert.get(alert.alert_id, as_obj=False)['metadata']['replay'] == 'done'
    assert os.path.exists(filename)

    # Delete the alert to test the loading process
    datastore.alert.delete(alert.alert_id)
    datastore.alert.commit()

    # In case the replay.yaml config creator output is not the same as loader input
    new_filename = filename.replace(output_dir, input_dir)
    if filename != new_filename:
        os.rename(filename, new_filename)
        filename = new_filename

    # Test replay loader
    loader.load_files(once=True)
    assert loader.client.queues['file'].length() == 1
    assert loader.client.queues['file'].peek_next() == filename

    # Test replay loader worker
    loader_worker.process_file(once=True)
    assert loader_worker.client.queues['file'].length() == 0
    assert not os.path.exists(filename)

    loaded_alert = datastore.alert.get(alert.alert_id, as_obj=False)
    assert 'bundle.loaded' in loaded_alert['metadata']
    assert alert.alert_id == loaded_alert['alert_id']
    assert alert.workflows_completed != loaded_alert['workflows_completed']


def test_replay_single_submission(config, datastore, creator, creator_worker, loader, loader_worker):
    output_dir = creator.replay_config.creator.output_filestore.replace('file://', '')
    input_dir = loader.replay_config.loader.input_directory

    # Make sure the submission get picked up by the creator
    sub = random.choice(all_submissions).as_primitives()
    sub['metadata']['replay'] = 'requested'
    datastore.submission.save(sub['sid'], sub)
    datastore.submission.commit()
    filename = os.path.join(output_dir, f"submission_{sub['sid']}.al_bundle")

    # Test replay creator
    creator.client.setup_submission_input_queue(once=True)
    assert creator.client.queues['submission'].length() == 1
    assert creator.client.queues['submission'].peek_next()['sid'] == sub['sid']

    # Test replay creator worker
    creator_worker.process_submission(once=True)
    datastore.submission.commit()
    assert creator_worker.client.queues['submission'].length() == 0
    assert datastore.submission.get(sub['sid'], as_obj=False)['metadata']['replay'] == 'done'
    assert os.path.exists(filename)

    # Delete the alert to test the loading process
    datastore.submission.delete(sub['sid'])
    datastore.submission.commit()

    # In case the replay.yaml config creator output is not the same as loader input
    new_filename = filename.replace(output_dir, input_dir)
    if filename != new_filename:
        os.rename(filename, new_filename)
        filename = new_filename

    # Test replay loader
    loader.load_files(once=True)
    assert loader.client.queues['file'].length() == 1
    assert loader.client.queues['file'].peek_next() == filename

    # Test replay loader worker
    loader_worker.process_file(once=True)
    assert loader_worker.client.queues['file'].length() == 0
    assert not os.path.exists(filename)

    loaded_submission = datastore.submission.get(sub['sid'], as_obj=False)
    assert 'bundle.loaded' in loaded_submission['metadata']
    assert sub['sid'] == loaded_submission['sid']
    assert 'replay' not in loaded_submission['metadata']


@pytest.mark.parametrize("collection", ["badlist", "safelist", "signature", "workflow"])
def test_replay_single_data_collection(datastore, creator, creator_worker, loader, loader_worker, collection):
    output_dir = creator.replay_config.creator.output_filestore.replace('file://', '')
    input_dir = loader.replay_config.loader.input_directory

    # Make sure the item gets picked up by the creator
    item = random.choice(data_collections[collection])

    # Test replay creator
    getattr(creator.client, f'setup_{collection}_input_queue')(once=True)
    assert creator.client.queues[collection].length() == 1
    assert creator.client.queues[collection].peek_next()['id'] == item.id
    # Test replay creator worker
    getattr(creator_worker, f'process_{collection}')(once=True)
    assert creator_worker.client.queues[collection].length() == 0
    filename = os.path.join(output_dir,
                            ([f for f in os.listdir(output_dir) if f.startswith(
                                collection) and f.endswith('.al_json')] + ["not_found"])[0])
    assert os.path.exists(filename)

    # Delete the item to test the loading process
    getattr(datastore, collection).delete(item.id)
    getattr(datastore, collection).commit()

    # In case the replay.yaml config creator output is not the same as loader input
    new_filename = filename.replace(output_dir, input_dir)
    if filename != new_filename:
        os.rename(filename, new_filename)
        filename = new_filename

    # Test replay loader
    loader.load_files(once=True)
    assert loader.client.queues['file'].length() == 1
    assert loader.client.queues['file'].peek_next() == filename

    # Test replay loader worker
    loader_worker.process_file(once=True)
    assert loader_worker.client.queues['file'].length() == 0
    assert not os.path.exists(filename)

    loaded_item = getattr(datastore, collection).get_if_exists(item.id)
    assert loaded_item == item

    if collection == "workflow":
        # Test updating the workflow but by a different author (reverse the author's name) and change the enabled state
        item.edited_by = item.edited_by[::-1]
        item.enabled = not item.enabled

        new_workflow_fn = os.path.join(input_dir, "workflow_blah.al_json")
        with open(new_workflow_fn, 'w') as fp:
            json_blob = {"id": item.id}
            json_blob.update(item.as_primitives())
            json.dump([json_blob], fp)

        # Load file to be processed
        loader.load_files(once=True)
        loader_worker.process_file(once=True)

        # If the workflow was edited by someone else, they shouldn't have control over the enabled state
        loaded_workflow = datastore.workflow.get(item.workflow_id)
        assert loaded_workflow.enabled != item.enabled and loaded_workflow.edited_by == item.edited_by
